import unittest
try:
    from unittest import mock
except ImportError:
    # Python 3.2 does not have mock in the standard library
    import mock

from sauna.plugins import (human_to_bytes, bytes_to_human, Plugin,
                           PluginRegister)
from sauna.plugins.ext import puppet_agent, postfix, memcached


class PluginsTest(unittest.TestCase):

    def test_human_to_bytes(self):
        self.assertEquals(0, human_to_bytes('0'))
        self.assertEquals(1000, human_to_bytes('1000'))
        self.assertEquals(1024, human_to_bytes('1K'))
        self.assertEquals(1024, human_to_bytes('1KiB'))
        self.assertEquals(1024, human_to_bytes('1KiO'))
        self.assertEquals(5242880, human_to_bytes('5M'))
        self.assertEquals(10737418240, human_to_bytes('10G'))

    def test_bytes_to_human(self):
        self.assertEquals('0B', bytes_to_human(0))
        self.assertEquals('1000B', bytes_to_human(1000))
        self.assertEquals('1.0K', bytes_to_human(1024))
        self.assertEquals('5.0M', bytes_to_human(5*1024*1024))
        self.assertEquals('10.0G', bytes_to_human(10*1024*1024*1024))

    def test_strip_percent_sign(self):
        self.assertEquals(10, Plugin._strip_percent_sign('10'))
        self.assertEquals(10, Plugin._strip_percent_sign('10%'))
        self.assertEquals(-10, Plugin._strip_percent_sign('-10%'))

    def test_strip_percent_sign_from_check_config(self):
        check_config = {
            'warn': '75%',
            'crit': 95,
            'something_random': '10'
        }
        expected_config = {
            'warn': 75,
            'crit': 95,
            'something_random': '10'
        }
        self.assertDictEqual(
            expected_config,
            Plugin._strip_percent_sign_from_check_config(check_config)
        )

    def test_get_threshold(self):
        check_config = {
            'warn': 75,
            'crit': 95,
            'something_random': '10'
        }

        # Test without modifier
        self.assertTupleEqual((95, 75), Plugin.get_thresholds(check_config))

        # Test with modifier
        def double(number):
            return number * 2
        self.assertTupleEqual(
            (190, 150),
            Plugin.get_thresholds(check_config, double)
        )

    def test_value_to_status(self):
        less = Plugin._value_to_status_less
        more = Plugin._value_to_status_more

        # Test less, imagine this check is about used RAM in percent
        check_config = {
            'warn': 75,
            'crit': 95,
        }
        self.assertEquals(Plugin.STATUS_OK, less(10, check_config))
        self.assertEquals(Plugin.STATUS_WARN, less(80, check_config))
        self.assertEquals(Plugin.STATUS_CRIT, less(100, check_config))

        # Test less, imagine this check is about free RAM in percent
        check_config = {
            'warn': 25,
            'crit': 5,
        }
        self.assertEquals(Plugin.STATUS_OK, more(50, check_config))
        self.assertEquals(Plugin.STATUS_WARN, more(20, check_config))
        self.assertEquals(Plugin.STATUS_CRIT, more(2, check_config))

    def test_get_all_plugins(self):
        plugins = PluginRegister.all_plugins
        self.assertIsInstance(plugins, dict)
        self.assertGreater(len(plugins), 1)
        for plugin_name, plugin_info in plugins.items():
            self.assertIn('plugin_cls', plugin_info)
            self.assertIn('checks', plugin_info)
            self.assert_(issubclass(plugin_info['plugin_cls'], Plugin))
            self.assertIsInstance(plugin_info['checks'], dict)

    def test_get_plugin(self):
        import sauna
        sauna.Sauna.import_submodules('sauna.plugins.ext')
        load_plugin = PluginRegister.get_plugin('Load')
        self.assert_(issubclass(load_plugin['plugin_cls'], Plugin))
        self.assertIsNone(PluginRegister.get_plugin('Unknown'))


class PuppetAgentTest(unittest.TestCase):

    def setUp(self):
        self.agent = puppet_agent.PuppetAgent({})

    @mock.patch('sauna.plugins.ext.puppet_agent.time')
    def test_last_run_delta(self, mock_time):
        mock_time.time.return_value = 1460836285.4
        self.agent._last_run_summary = {
            'time': {
                'last_run': 1460836283
            }
        }
        self.assertTupleEqual(
            self.agent.last_run_delta({'warn': 10, 'crit': 20}),
            (Plugin.STATUS_OK, 'Puppet last ran 0:00:02 ago')
        )

        mock_time.time.return_value = 1460836295.4
        self.assertTupleEqual(
            self.agent.last_run_delta({'warn': 10, 'crit': 20}),
            (Plugin.STATUS_WARN, 'Puppet last ran 0:00:12 ago')
        )

    def test_failures(self):
        self.agent._last_run_summary = {
            'events': {
                'failure': 0
            }
        }
        self.assertTupleEqual(
            self.agent.failures({'warn': 1, 'crit': 2}),
            (Plugin.STATUS_OK, 'Puppet ran without trouble')
        )

        self.agent._last_run_summary['events']['failure'] = 1
        self.assertTupleEqual(
            self.agent.failures({'warn': 1, 'crit': 1}),
            (Plugin.STATUS_CRIT,
             'Puppet last run had 1 failure(s)')
        )

    @mock.patch('builtins.open')
    def test_get_last_run_summary(self, mock_open):
        class ContextManager:
            def __enter__(self):
                return '''
---
  time:
    last_run: 1460836283
  events:
    failure: 0
    success: 1
    total: 1
'''

            def __exit__(self, exc_type, exc_val, exc_tb):
                pass

        mock_open.return_value = ContextManager()
        expected = {
            'time': {
                'last_run': 1460836283
            },
            'events': {
                'failure': 0,
                'success': 1,
                'total': 1,
            }
        }
        self.assertEquals(self.agent._last_run_summary, None)
        self.assertDictEqual(self.agent.last_run_summary, expected)
        self.assertDictEqual(self.agent._last_run_summary, expected)


class PostfixTest(unittest.TestCase):

    def setUp(self):
        self.postfix = postfix.Postfix({})

    def test_get_queue_size(self):
        self.postfix._mailq_output = 'Mail queue is empty\n'
        self.assertEqual(self.postfix._get_queue_size(), 0)

        self.postfix._mailq_output = (
            'postqueue: fatal: Queue report unavailable '
            '- mail system is down\n'
        )
        with self.assertRaises(Exception):
            self.postfix._get_queue_size()

        self.postfix._mailq_output = '''
        -Queue ID- --Size-- ----Arrival Time---- -Sender/Recipient-------
89B3CC02*       436 Tue Apr 19 09:21:31  user@host
                                         user2@host

-- 0 Kbytes in 1 Request.
'''
        self.assertEqual(self.postfix._get_queue_size(), 1)

        self.postfix._mailq_output = '-- 105 Kbytes in 25 Requests.\n'
        self.assertEqual(self.postfix._get_queue_size(), 25)

    def test_queue_size(self):
        self.postfix._get_queue_size = lambda: 5
        self.assertTupleEqual(
            self.postfix.queue_size({'warn': 10, 'crit': 20}),
            (Plugin.STATUS_OK, '5 mail(s) in queue')
        )
        self.assertTupleEqual(
            self.postfix.queue_size({'warn': 1, 'crit': 20}),
            (Plugin.STATUS_WARN, '5 mail(s) in queue')
        )
        self.assertTupleEqual(
            self.postfix.queue_size({'warn': 1, 'crit': 2}),
            (Plugin.STATUS_CRIT, '5 mail(s) in queue')
        )


class MemcachedTest(unittest.TestCase):

    def setUp(self):
        self.memcached = memcached.Memcached({})

    def test_bytes(self):
        self.memcached._stats = {'bytes': 1024}
        self.assertTupleEqual(
            self.memcached.bytes({'warn': '2K', 'crit': '4K'}),
            (Plugin.STATUS_OK, 'Memcached memory: 1.0K')
        )
        self.assertTupleEqual(
            self.memcached.bytes({'warn': '512', 'crit': '4K'}),
            (Plugin.STATUS_WARN, 'Memcached memory: 1.0K')
        )
        self.memcached._stats = {'bytes': 100000}
        self.assertTupleEqual(
            self.memcached.bytes({'warn': '512', 'crit': '4K'}),
            (Plugin.STATUS_CRIT, 'Memcached memory: 97.7K')
        )

    def test_used_percent(self):
        self.memcached._stats = {'bytes': 512, 'limit_maxbytes': 1024}
        self.assertTupleEqual(
            self.memcached.used_percent({'warn': '40%', 'crit': '80%'}),
            (Plugin.STATUS_WARN, 'Memcached memory used: 50% of 1.0K')
        )

    def test_current_items(self):
        self.memcached._stats = {'curr_items': 100}
        self.assertTupleEqual(
            self.memcached.current_items({'warn': 50, 'crit': 200}),
            (Plugin.STATUS_WARN, 'Memcached holds 100 items')
        )

    def test_accepting_connections(self):
        self.memcached._stats = {'accepting_conns': 1}
        self.assertTupleEqual(
            self.memcached.accepting_connections({}),
            (Plugin.STATUS_OK, 'Memcached is accepting connections')
        )
        self.memcached._stats = {'accepting_conns': 0}
        self.assertTupleEqual(
            self.memcached.accepting_connections({}),
            (Plugin.STATUS_CRIT, 'Memcached is not accepting connections')
        )

    def test_raw_stats_to_dict(self):
        data = b'''STAT pid 8509
STAT accepting_conns 1
STAT bytes 70
STAT curr_items 1
END
'''
        self.assertDictEqual(
            self.memcached._raw_stats_to_dict(data),
            {
                'pid': 8509,
                'accepting_conns': 1,
                'bytes': 70,
                'curr_items': 1
            }
        )
